from farmi.farmi import Farmi
import json
from collections import defaultdict
import zmq
import re
import msgpack


class Subscriber(Farmi):
    def __init__(self, directory_service_address='tcp://127.0.0.1:5555'):
        super().__init__(directory_service_address)
        self.topics = defaultdict(dict)
        self.poller = zmq.Poller()

        self.directory_service.send_json({'action': 'GET_PUB_ADDRESS'})
        sub_address = self.directory_service.recv_string()

        self.directory_service_sub = self.context.socket(zmq.SUB)
        self.directory_service_sub.connect(sub_address)
        self.directory_service_sub.subscribe('')

        self.poller.register(self.directory_service_sub, zmq.POLLIN)

    def get_matching_topics(self, topic):
        for topic_matchers in self.topics.keys():
            if topic_matchers.match(topic):
                yield topic_matchers

    def subscribe_to(self, topic, fn):
        try:
            topic.match('')
        except:
            topic = re.compile('^%s$' % re.escape(topic))
        self.topics[topic]['fn'] = fn

        self.directory_service.send_json({'action': 'TOPICS'})
        topics = self.directory_service.recv_json()
        for received_topic, info in topics.items():
            for matching_topic in self.get_matching_topics(received_topic):
                self._add_topic(matching_topic, info['address'])

    def _add_topic(self, topic, address):
        self.topics[topic]['address'] = address
        socket = self.context.socket(zmq.SUB)
        socket.connect(address)
        socket.subscribe('')
        self.topics[topic]['socket'] = socket
        self.poller.register(self.topics[topic]['socket'], zmq.POLLIN)

    def _remove_topic(self, topic):
        if self.topics[topic].get('socket'):
            self.poller.unregister(self.topics[topic]['socket'])
            self.topics[topic]['socket'].close()
            self.topics[topic].pop('socket')
            self.topics[topic].pop('address')
            

    def close(self):
        for topic, info in self.topics.items():
            if info['socket']:
                info['socket'].close()
        super().close()

    def listen(self):
        while not self.exit.is_set():
            poller = dict(self.poller.poll())
            for poll in poller:
                if poll == self.directory_service_sub:
                    _, raw_msg = poll.recv_multipart()
                    msg = json.loads(raw_msg.decode('utf-8'))
                    action = msg.get('action')
                    if action == 'REGISTERED':
                        topic, address = msg.get('topic'), msg.get('address')
                        for matching_topic in self.get_matching_topics(topic):
                            if self.topics[matching_topic].get('address') != address:
                                self._add_topic(matching_topic, address)
                    elif action == 'DEREGISTERED':
                        for matching_topic in list(self.get_matching_topics(msg.get('topic')))[:]:
                            self._remove_topic(matching_topic)
                    for matching_topic in self.get_matching_topics(action):
                        self.topics[matching_topic]['fn'](msg.get('topic'), msg.get('time'), msg)
                else:
                    for topic in self.topics.values():
                        if topic.get('socket') == poll:
                            msg = poll.recv_multipart()
                            if len(msg) == 3:
                                try:
                                    topic_, time_, body = msg
                                except ValueError as e:
                                    print(msg)
                                    raise e
                                topic['fn'](topic_.decode('utf-8'), float(time_.decode('utf-8')), msgpack.unpackb(body, raw=False))



import time
from threading import Thread
import msgpack
import msgpack_numpy as m
m.patch()
import zmq
from farmi.utils import get_ip
from farmi.farmi import Farmi


class Publisher(Farmi):
    def __init__(self, topic, local_save=None, directory_service_address='tcp://127.0.0.1:5555', heartbeat_frequency=5):
        super().__init__(directory_service_address)
        self.topic = topic
        self.heartbeat_frequency = heartbeat_frequency
        self.pub_address = None
        self.pub_socket = self.context.socket(zmq.PUB)
        
        self._create_publisher()
        Thread(target=self._heartbeat).start()
        
        if local_save:
            self.packer = msgpack.Packer()
            dir_name = local_save if isinstance(local_save, str) else '.'
            timestamp = time.strftime("%Y_%m_%d_%H_%M_%S")
            self.file_handle = open('{}/{}-{}.farmi'.format(dir_name, timestamp, self.topic), 'wb')
        else:
            self.file_handle = None

    def _heartbeat(self):
        while not self.exit.is_set():
            self.directory_service.send_json({
                'action': 'HEARTBEAT',
                'topic': self.topic,
                'time': time.time()
            })
            response = self.directory_service.recv_string()
            if response == 'NOT_REGISTERED':
                self._register()
            self.exit.wait(self.heartbeat_frequency)

    def _register(self):
        self.directory_service.send_json({
            'action': 'REGISTER',
            'topic': self.topic,
            'address': self.pub_address
        })
        self.directory_service.recv_string()

    def _create_publisher(self):
        zmq_port = self.pub_socket.bind_to_random_port('tcp://*', max_tries=150)
        self.pub_address = 'tcp://{}:{}'.format(get_ip(), zmq_port)
        self._register()


    def get_shifted_time(self):
        return time.time() + self.time_offset

    def send(self, data, timestamp=None):
        if not timestamp:
            timestamp = time.time()
        data_time = timestamp + self.time_offset
        if self.file_handle:
            self.file_handle.write(self.packer.pack((self.topic, data_time, data)))
            self.file_handle.flush()
        self.pub_socket.send_multipart([self.topic.encode('utf-8'), str(data_time).encode('utf-8'), msgpack.packb(data, use_bin_type=True)])

    def close(self):
        self.directory_service.send_json({
            'action': 'DEREGISTER',
            'topic': self.topic
        })
        self.directory_service.recv()
        super().close()
        self.pub_socket.send(b'CLOSE')
        self.pub_socket.close()
        if self.file_handle:
            self.file_handle.close()



import time
import zmq
import json
from farmi.utils import get_ip
import logging

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class DirectoryService(object):
    def __init__(self, port=5555, purge_timeout=20):
        self._topics = {}
        self.purge_timeout = purge_timeout
        context = zmq.Context()
        self.rep_socket = context.socket(zmq.REP)
        self.rep_socket.bind('tcp://*:{}'.format(port))

        self.publisher = context.socket(zmq.PUB)
        pub_port = self.publisher.bind_to_random_port('tcp://*', max_tries=150)
        self.pub_address = 'tcp://{}:{}'.format(get_ip(), pub_port)

    @property
    def topics(self):
        self._purge_topics()
        return self._topics

    def _purge_topics(self):
        to_be_deleted = []
        for topic, info in self._topics.items():
            if time.time() - info['latest_heartbeat'] > self.purge_timeout:
                to_be_deleted.append(topic)

        for topic in to_be_deleted:
            self._deregister(topic)

    def _deregister(self, topic):
        if self._topics.get(topic):
            self._topics.pop(topic)
            self._broadcast(json.dumps({
                'action': 'DEREGISTERED',
                'topic': topic
            }))

    def listen(self):
        while True:
            raw_msg = self.rep_socket.recv_string()
            try:
                msg = json.loads(raw_msg)
            except ValueError:
                logger.info('MALFORMED_JSON' + raw_msg)
                self.rep_socket.send_string('MALFORMED_JSON')
                continue

            action = msg.get('action')

            if action == 'GET_PUB_ADDRESS':
                response = self.handle_get_pub_address(msg)
            elif action == 'REGISTER':
                response = self.handle_register(msg)
            elif action == 'DEREGISTER':
                response = self.handle_deregister(msg)
            elif action == 'SYNC':
                response = self.handle_sync(msg)
            elif action == 'ANNOUNCE':
                response = self.handle_announce(msg)
            elif action == 'HEARTBEAT':
                response = self.handle_heartbeat(msg)
            elif action == 'TOPICS':
                response = self.handle_topics(msg)
            else:
                response = 'COMMAND_NOT_FOUND'

            logger.info('received msg: {}, responded: {}'.format(msg, response))
            self.rep_socket.send_string(response)

    def _broadcast(self, data, topic=''):
        logger.info('broadcasting: {}'.format(data))
        self.publisher.send_multipart([topic.encode('utf-8'), data.encode('utf-8')])

    def _check_required_args(self, msg, required):
        return all([msg.get(x) for x in required])

    def handle_topics(self, msg):
        return json.dumps(self.topics)

    def handle_get_pub_address(self, msg):
        return self.pub_address

    def handle_deregister(self, msg):
        if not self._check_required_args(msg, ['topic']):
            return 'MISSING_INFORMATION'
        self._deregister(msg.get('topic'))
        return 'OK'

    def handle_register(self, msg):
        if not self._check_required_args(msg, ['topic', 'address']):
            return 'MISSING_INFORMATION'

        topic, address = msg.get('topic'), msg.get('address')

        if self.topics.get(topic):
            self._deregister(topic)
            # return 'TOPIC_ALREADY_REGISTERED'

        self.topics[topic] = {
            'address': address,
            'latest_heartbeat': time.time()
        }
        self._broadcast(json.dumps({
            'action': 'REGISTERED',
            'topic': topic,
            'address': address
        }))
        return 'OK'

    def handle_sync(self, msg):
        return str(time.time())

    def handle_announce(self, msg):
        if not self._check_required_args(msg, ['msg']):
            return 'MISSING_INFORMATION'

        self._broadcast(msg.get('msg'))
        return 'OK'

    def handle_heartbeat(self, msg):
        if not self._check_required_args(msg, ['topic', 'time']):
            return 'MISSING_INFORMATION'

        logger.info('got HEARTBEAT: {}'.format(msg))

        topic, time_ = msg.get('topic'), msg.get('time')

        if self._topics.get(topic):
            self._topics[topic]['latest_heartbeat'] = time.time()

            self._broadcast(json.dumps({
                'action': 'HEARTBEAT',
                'topic': topic,
                'time': time.time(),
                'sender_time': time_
            }))
            return 'OK'
        else:
            return 'NOT_REGISTERED'



#!/usr/bin/env python
from setuptools import setup, find_packages


setup(
    name='farmi',
    install_requires=[
        'pyzmq',
        'msgpack',
        'msgpack-numpy'
    ],
    python_requires='>=3',
    version='3.0.7',
    description='Multisensory processing framework ',
    author='Patrik Jonell',
    author_email='pjjonell@kth.se',
    packages=find_packages(),
    scripts=[
        'bin/farmi-server'
    ],
)



import socket


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip
